from unittest import TestCase
from ..txtui import show


class TxtUiTests(TestCase):
    """
    Some simple tests for rendering values.
    """

    def test_show(self) -> None:
        """
        L{show} renders a password prompt which includes placeholders both for
        hidden and forgotten elements of the underlying generated password.
        """

        self.assertEqual(
            show("/", ["two", "three", "four"], 4, 1), "••••/°°°°/three/four"
        )
        self.assertEqual(
            show("/", ["two", "three", "four"], 4, 2), "••••/°°°°/°°°°/four"
        )



from __future__ import annotations

from dataclasses import dataclass
from secrets import token_bytes
from time import time
from typing import Any, Sequence

from .difficulty import SCryptParameters, determineScryptParameters, sysrand, oldDefaultScryptParams
from .txtui import promptUser


@dataclass
class Memorization:
    """
    A PIN or password whose memorization is in progress
    """

    label: str
    """
    The name of this new PIN.
    """

    remainingTokens: list[str]
    """
    The digits or words we haven't memorized yet.
    """

    tokensMemorized: int
    """
    The number of digits or words the user has already memorized.
    """

    successCount: int
    """
    How many times in a row the user has succeeded at success.
    """

    separator: str
    """
    The separator between characters.
    """

    salt: bytes
    """
    Salt for deriving the correct hash.
    """

    key: bytes
    """
    The result of hashing the password being memorized.
    """

    entryTimes: list[tuple[float, bool]]

    kdf: SCryptParameters

    @property
    def done(self) -> bool:
        return len(self.remainingTokens) == 0

    @classmethod
    def new(
        cls,
        label: str,
        tokens: Sequence[str] = "1234567890",
        length: int = 6,
        separator: str = "",
    ) -> Memorization:
        """
        create a new password to memorize
        """
        remainingTokens = [sysrand.choice(tokens) for _ in range(length)]
        salt = token_bytes(16)
        kdf = determineScryptParameters()
        password = separator.join(remainingTokens).encode("utf-8")
        key = kdf.kdf(salt=salt, password=password)
        return Memorization(
            label=label,
            remainingTokens=remainingTokens,
            tokensMemorized=0,
            successCount=0,
            separator=separator,
            salt=salt,
            key=key,
            kdf=kdf,
            entryTimes=[],
        )

    def string(self) -> str:
        groupSeparator = "/"
        placeholderChar = "•"
        placeholder: str = placeholderChar * 4 if self.separator else placeholderChar
        allTokens = self.remainingTokens + (self.tokensMemorized * [placeholder])
        allTokens.insert(
            (len(self.remainingTokens) + self.tokensMemorized) // 2, groupSeparator
        )
        return self.separator.join(allTokens)

    def prompt(self) -> bool:
        correct = promptUser(
            nextTime=self.nextPromptTime(),
            label=self.label,
            kdf=self.kdf,
            salt=self.salt,
            key=self.key,
            separator=self.separator,
            knownTokens=self.remainingTokens,
            totalTokens=len(self.remainingTokens) + self.tokensMemorized,
            hiddenTokens=0,
        )
        if correct is None:
            return False
        self.entryTimes.append((time(), correct))
        if correct:
            SUCCESS_THRESHOLD = 5
            self.successCount += 1
            print(f"✅ Yay, correct {self.successCount}/{SUCCESS_THRESHOLD} times")
            if self.successCount >= SUCCESS_THRESHOLD and self.remainingTokens:
                self.tokensMemorized += 1
                self.remainingTokens.pop(-1)
                print("🎉 Level Up! 🎊")
                self.successCount = 0
            return True
        else:
            self.successCount = 0
            print("❌ Oops, try again")
            return False

    def tojson(self) -> dict[str, object]:
        """
        convert to json-serializable dict
        """
        return {
            "label": self.label,
            "remainingTokens": self.remainingTokens,
            "tokensMemorized": self.tokensMemorized,
            "successCount": self.successCount,
            "separator": self.separator,
            "salt": self.salt.hex(),
            "key": self.key.hex(),
            "entryTimes": self.entryTimes,
            "kdf": self.kdf.tojson(),
        }

    @classmethod
    def fromjson(cls, data: dict[str, Any]) -> Memorization:
        """
        convert from json-serializable dict
        """
        return Memorization(
            label=data["label"],
            remainingTokens=data["remainingTokens"],
            tokensMemorized=data["tokensMemorized"],
            successCount=data["successCount"],
            separator=data["separator"],
            salt=bytes.fromhex(data["salt"]),
            key=bytes.fromhex(data["key"]),
            entryTimes=data["entryTimes"],
            kdf=SCryptParameters.fromjson(data.get("kdf", oldDefaultScryptParams)),
        )

    def nextPromptTime(self) -> float:
        """
        When should we next prompt the user?
        """
        if not self.entryTimes:
            # I've never guessed; let's guess right now.
            return time()
        timestamp, correct = self.entryTimes[-1]
        if not correct:
            return time()
        # need to delay. want to memorize a password in around 3 days or so. 6
        # digits, 5 correct guesses per digit necessary.  30 guesses minimum.
        return timestamp + min(86400, (90 * (2**self.successCount)))



from os.path import exists
from time import time

from .app import load, timecache, PinPalApp
from .mem1 import Memorization
from .mem2 import Memorization2


def doSelfTest() -> None:
    from json import dumps, loads
    testing: Memorization2 | Memorization = Memorization2.new("testing")
    while True:
        testing = load(loads(dumps(testing.tojson())))
        testing.prompt()


def main() -> None:
    """
    Run the tool.
    """
    from sys import argv, exit, stdout

    if len(argv) > 1 and argv[1] == "check":
        if exists(timecache):
            with open(timecache) as f:
                needsCheckAt = float(f.read())
            if needsCheckAt < time():
                stdout.write(" 📌⏰")
        exit(0)

    subCommand = None if len(argv) < 2 else argv[1]

    if subCommand == "test":
        doSelfTest()

    app = (
        PinPalApp([])
        if (subCommand == "clear") or (maybeApp := PinPalApp.load()) is None
        else maybeApp
    )

    if subCommand == "new":
        newLabel = input("What do you want to call this new PIN?")
        m = Memorization2.new(newLabel)
        app.memorizations.append(m)

    elif subCommand == "list":
        for idx, mem in enumerate(app.memorizations):
            print(f"{idx}: {mem.label} {'done' if mem.done else 'in-progress'}")

    elif subCommand == "drop":
        for idx, mem in enumerate(app.memorizations):
            print(f"{idx}: {mem.label}")
        dropnum = input("Which number do you want to drop? ")
        dropidx = int(dropnum)
        if (input(f"Dropping {mem.label}, OK?")) == "yes":
            del app.memorizations[dropidx]
            print("OK, dropped.")

    else:
        for each in app.memorizations:
            each.prompt()
    app.save()






from __future__ import annotations

from dataclasses import dataclass
from hashlib import scrypt
from secrets import SystemRandom, token_bytes
from time import time

oldDefaultScryptParams = {
    "n": str(0x4000),
    "r": str(8),
    "p": str(1),
    "maxmem": str(64 * 1024 * 1024),
}

sysrand = SystemRandom()

@dataclass
class SCryptParameters:
    """
    Keyword parameters for L{scrypt}.
    """

    r: int
    p: int
    n: int
    maxmem: int

    def kdf(self, *, salt: bytes, password: bytes) -> bytes:
        return scrypt(
            password, salt=salt, r=self.r, p=self.p, n=self.n, maxmem=self.maxmem
        )

    @classmethod
    def fromjson(cls, json: dict[str, str]) -> SCryptParameters:
        """
        Load SCrypt parameters from some serialized JSON objects.
        """
        return cls(
            r=int(json["r"]),
            p=int(json["p"]),
            n=int(json["n"]),
            maxmem=int(json["maxmem"]),
        )

    def tojson(self) -> dict[str, str]:
        """
        Convert SCrypt parameters to JSON.
        """
        return {
            "r": str(self.r),
            "p": str(self.p),
            "n": str(self.n),
            "maxmem": str(self.maxmem),
        }


determined: SCryptParameters | None = None


def determineScryptParameters(times: int = 10) -> SCryptParameters:
    """
    Determine an ideal value for `n` and `maxmem`, per U{this comment,
    <https://go-review.googlesource.com/c/crypto/+/67070/3/scrypt/scrypt.go#223>}

    'consider setting N to the highest power of 2 you can derive within 100
    milliseconds'
    """
    global determined
    if determined is not None:
        return determined
    salt = token_bytes(16)
    password = token_bytes(16)
    r = 8
    p = 1
    nPower = 13
    n = 1 << nPower

    while True:
        then = time()
        previousN = n
        n = 1 << nPower
        # documented in Node, but not Python, apparently: “It is an error when
        # (approximately) 128 * N * r > maxmem. Default: 32 * 1024 * 1024.”
        # https://nodejs.org/api/crypto.html#cryptoscryptsyncpassword-salt-keylen-options
        maxmem = 128 * n * r * 2
        # '* 2' added on the end here because we stil seem to bump into memory
        # issues when set to exactly 128*n*r
        for _ in range(times):
            scrypt(salt=salt, password=password, r=r, p=p, n=n, maxmem=maxmem)

        now = time()
        if ((now - then) / times) > 0.1:
            determined = SCryptParameters(r=r, p=p, n=previousN, maxmem=maxmem)
            return determined
        nPower += 1



# -*- test-case-name: pinpal.test.test_txtui -*-
from .difficulty import SCryptParameters
from time import time
from getpass import getpass


def promptUser(
    *,
    nextTime: float,
    label: str,
    kdf: SCryptParameters,
    salt: bytes,
    key: bytes,
    separator: str,
    knownTokens: list[str],
    totalTokens: int,
    hiddenTokens: int,
    forgottenChar: str = "•",
    hiddenChar: str = "°",
    attempts: int = 4,
) -> bool | None:
    """
    Prompt the user.
    """

    remaining = nextTime - time()
    if remaining > 0:
        print("next reminder for", label, "in", int(remaining), "seconds")
        return None
    attempt = ""
    for repetition in range(attempts):
        reshow = show(
            separator,
            knownTokens,
            totalTokens,
            hiddenTokens,
            forgottenChar,
            hiddenChar,
        )
        userInput = getpass(f"\n\n\n{label} (reminder: {reshow}){attempt}: ")
        attempt = f" (attempt {repetition + 2}/{attempts})"
        if kdf.kdf(salt=salt, password=userInput.encode("utf-8")) == key:
            return True
        hiddenTokens = 0
    return False


def show(
    separator: str,
    knownTokens: list[str],
    totalTokens: int,
    hiddenTokens: int,
    forgottenChar: str = "•",
    hiddenChar: str = "°",
) -> str:
    """
    Show a partially-obscured passphrase based on a list of tokens that are
    still stored in plaintext, a total number of tokens, and a placeholder.
    """

    # TODO: placeholder styling ought to be the responsibility of TokenType,
    # with •••• being TokenType.words and • being TokenType.numbers.
    def inflate(ch: str) -> str:
        return ch * 4 if separator else ch

    forgottenPlaceholder: str = inflate(forgottenChar)
    hiddenPlaceholder: str = inflate(hiddenChar)
    allTokens = (
        ((totalTokens - len(knownTokens)) * [forgottenPlaceholder])
        + ([hiddenPlaceholder] * hiddenTokens)
        + knownTokens[hiddenTokens:]
    )
    return separator.join(allTokens)



from objc import object_property
from AppKit import NSApplication, NSNib, NSTableView, NSTableColumn
from Foundation import NSObject
from quickmacapp import Status, mainpoint  # , answer

# from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IReactorTime

from .app import PinPalApp
from .mem2 import Memorization2


class MemorizationDataSource(NSObject):
    pinPalApp: PinPalApp = object_property()
    selectedRow: NSObject = object_property()

    def awakeFromNib(self) -> None:
        loaded = PinPalApp.load()
        if loaded is None:
            loaded = PinPalApp([])
        self.pinPalApp = loaded

    def tableViewSelectionDidChange_(self, notification: NSObject) -> None:
        self.selectedRow = self.tableView_objectValueForTableColumn_row_(
            None, None, notification.object().selectedRowIndexes().firstIndex()
        )

    def numberOfRowsInTableView_(
        self,
        tableView: NSTableView,
    ) -> int:
        return len(self.pinPalApp.memorizations)

    def tableView_objectValueForTableColumn_row_(
        self,
        tableView: NSTableView,
        column: NSTableColumn,
        row: int,
    ) -> object:
        item = self.pinPalApp.memorizations[row]
        return {
            "label": item.label,
            "guesses": (
                len(item.guesses)
                if isinstance(item, Memorization2)
                else item.successCount
            ),
        }


class PINPalAppOwner(NSObject):
    """
    NIB owner for the application.
    """


@mainpoint()
def main(reactor: IReactorTime) -> None:
    """
    Run oldMain by default so I can keep using the app while I'm working on a
    radical refactor of the object model in newMain.
    """
    status = Status("🔑🦃🗝")

    owner = PINPalAppOwner.alloc().init()

    def sayHello() -> None:
        # Deferred.fromCoroutine(answer("hi"))
        nibInstance = NSNib.alloc().initWithNibNamed_bundle_("PINList.nib", None)
        nibInstance.instantiateWithOwner_topLevelObjects_(owner, None)

    def bye() -> None:
        NSApplication.sharedApplication().terminate_(owner)

    sayHello()
    status.menu(
        [
            # ("Hello World", sayHello),
            ("Quit", bye),
        ]
    )



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Sequence
from time import time
from getpass import getpass
from secrets import token_bytes

from .difficulty import (
    SCryptParameters,
    oldDefaultScryptParams,
    determineScryptParameters,
    sysrand,
)
from .txtui import show, promptUser


@dataclass
class TokenInfo:
    tokens: Callable[[], Sequence[str]]
    separator: str
    retokenize: Callable[[str], Sequence[str]]


def horsephrase_tokens() -> Sequence[str]:
    from horsephrase.__main__ import words

    return words


class TokenType(Enum):
    numbers = TokenInfo(lambda: "0123456789", ", ", lambda x: list(x))
    words = TokenInfo(horsephrase_tokens, " ", lambda x: x.split(" "))


@dataclass
class UserGuess:
    """
    A user guessed the passphrase.
    """

    correct: bool
    """
    Was the user correct in their guess?
    """
    timestamp: float
    """
    What time was this done at?
    """
    length: int
    """
    How many tokens long was the thing they needed to guess?
    """

    def tojson(self) -> dict[str, object]:
        return {
            "correct": self.correct,
            "timestamp": self.timestamp,
            "length": str(self.length),
        }

    @classmethod
    def fromjson(cls, data: dict[str, Any]) -> UserGuess:
        return cls(
            correct=data["correct"],
            timestamp=data["timestamp"],
            length=int(data["length"]),
        )


@dataclass
class Memorization2:
    """
    Memorization stratetgy v2.0:

        - upon creation, remember what kind of secret we're generating, how
          many tokens we're targeting, and generate a single token.

        - when the user gets the guess correct N times, generate a new token,
          and save the hash of the new secret.

        - when the length of the secret gets over maxKnown tokens, forget the
          first token in the sequence.
    """

    label: str
    """
    The name of this new token we're generating.
    """

    targetTokenCount: int
    """
    The number of tokens we are ultimately hoping to generate.
    """

    knownTokens: list[str]
    """
    The list of tokens that the user is in the process of memorizing, that
    pinpal is storing in plaintext in order to prompt the user each time.
    """

    generatedCount: int
    """
    The total number of tokens generated so far by this memorization, including
    both plaintext-stored and already-forgotten tokens.
    """

    salt: bytes
    """
    Randomized salt for deriving the key. (16 random token bytes.)
    """

    key: bytes
    """
    The output of the KDF of all tokens generated so far (both plaintext-stored
    and already-forgotten).
    """

    tokenType: TokenType
    """
    The type of tokens being generated (i.e.: numbers L{TokenType.numbers} for
    a short PIN, words L{TokenType.words} for a passphrase).
    """

    guesses: list[UserGuess]
    """
    A record of all the user's guesses, both successful and unsuccessful.
    """

    maxKnown: int
    """
    The maximum number of tokens we can have stored in plaintext.
    """

    kdf: SCryptParameters
    """
    The parameters for the KDF.
    """

    @property
    def done(self) -> bool:
        """
        Have we generated every token that we are going to generate, and
        forgotten all plaintext tokens, storing only the metadata and KDF
        output?
        """
        return (self.generatedCount == self.targetTokenCount) and (
            0 == len(self.knownTokens)
        )

    def tojson(self) -> dict[str, object]:
        """
        convert to json-serializable dict
        """
        return {
            "label": self.label,
            "targetTokenCount": str(self.targetTokenCount),
            "knownTokens": self.knownTokens,
            "generatedCount": str(self.generatedCount),
            "salt": self.salt.hex(),
            "key": self.key.hex(),
            "tokenType": self.tokenType.name,
            "guesses": [each.tojson() for each in self.guesses],
            "maxKnown": str(self.maxKnown),
            "kdf": self.kdf.tojson(),
        }

    @classmethod
    def fromjson(cls, data: dict[str, Any]) -> Memorization2:
        """
        convert from json-serializable dict
        """
        return cls(
            label=data["label"],
            targetTokenCount=int(data["targetTokenCount"]),
            knownTokens=data["knownTokens"],
            generatedCount=int(data["generatedCount"]),
            salt=bytes.fromhex(data["salt"]),
            key=bytes.fromhex(data["key"]),
            tokenType=TokenType[data["tokenType"]],
            guesses=[UserGuess.fromjson(each) for each in data["guesses"]],
            maxKnown=int(data["maxKnown"]),
            kdf=SCryptParameters.fromjson(data.get("kdf", oldDefaultScryptParams)),
        )

    @classmethod
    def new(cls, label: str) -> Memorization2:
        """
        Create a new passphrase to memorize.
        """
        self = cls(
            label=label,
            targetTokenCount=5,
            knownTokens=[],
            generatedCount=0,
            salt=b"",
            key=b"",
            tokenType=TokenType.words,
            guesses=[],
            maxKnown=3,
            kdf=determineScryptParameters(),
        )
        self.generateOne()
        return self

    def string(self) -> str:
        return show(
            self.tokenType.value.separator,
            self.knownTokens,
            self.generatedCount,
            0 if not self.knownTokens else 1 if self.correctGuessCount() >= 2 else 0,
        )

    def nextPromptTime(self) -> float:
        """
        The time (in epoch seconds) at which the next prompt to the user ought
        to be displayed.
        """
        if not self.guesses:
            return time()
        else:
            return self.guesses[-1].timestamp + min(
                86400, (90 * (1.4 ** self.correctGuessCount()))
            )

    def correctThreshold(self) -> int:
        """
        How many guesses do we need to get correct in order to move on to the
        next token?
        """
        return (self.generatedCount) * 2

    def correctGuessCount(self) -> int:
        """
        Compute the number of continuous correct guesses at the current length
        of the password.
        """
        result = 0
        for each in reversed(self.guesses):
            if each.correct and each.length == self.generatedCount:
                result += 1
            else:
                return result
        return result

    def generateOne(self) -> None:
        """
        Generate one additional token.
        """
        chosen = sysrand.choice(self.tokenType.value.tokens())
        if self.generatedCount > self.maxKnown:
            # we no longer remember the entire passphrase, but the key has to
            # represent the entire passphrase.
            wholePassphrase = getpass(
                "enter correctly to confirm: "
                + show(
                    self.tokenType.value.separator,
                    self.knownTokens + [chosen],
                    self.generatedCount + 1,
                    0,
                )
                + ": "
            )
            tokens = self.tokenType.value.retokenize(wholePassphrase)
            newTokenMismatch = tokens[-1] != chosen
            oldTokenMismatch = (
                self.kdf.kdf(
                    salt=self.salt,
                    password=self.tokenType.value.separator.join(tokens[:-1]).encode(
                        "utf-8"
                    ),
                )
                != self.key
            )
            if newTokenMismatch or oldTokenMismatch:
                print("passphrase incorrect")
                return
        else:
            wholePassphrase = self.tokenType.value.separator.join(
                [*self.knownTokens, chosen]
            )
        # commit!
        self.generatedCount += 1
        self.knownTokens.append(chosen)
        if len(self.knownTokens) > self.maxKnown:
            self.knownTokens.pop(0)
        self.salt = token_bytes(16)
        self.key = self.kdf.kdf(
            salt=self.salt, password=wholePassphrase.encode("utf-8")
        )

    def prompt(self) -> bool:
        correct = promptUser(
            nextTime=self.nextPromptTime(),
            label=self.label,
            kdf=self.kdf,
            salt=self.salt,
            key=self.key,
            separator=self.tokenType.value.separator,
            knownTokens=self.knownTokens,
            totalTokens=self.generatedCount,
            hiddenTokens=0 if not self.knownTokens else 1 if self.correctGuessCount() >= 2 else 0,
        )
        if correct is None:
            return False
        self.guesses.append(
            UserGuess(correct=correct, timestamp=time(), length=self.generatedCount)
        )
        if correct:
            print("yay")
            guessesToGo = self.correctThreshold() - self.correctGuessCount()
            if guessesToGo > 0:
                print(guessesToGo, "correct entries to go before leveling up")
                return correct
            if not self.knownTokens:
                print("keep practicing!")
                return correct
            if self.generatedCount < self.targetTokenCount:
                print("leveling up")
                self.generateOne()
            else:
                print("forgetting some more")
                del self.knownTokens[0]
        else:
            print("too bad")
        return correct




from __future__ import annotations

from dataclasses import dataclass
from json import dumps, loads

from os.path import expanduser

from keyring import get_password, set_password

from .mem1 import Memorization
from .mem2 import Memorization2


timecache = expanduser("~/.pinpal-timestamp")


@dataclass
class PinPalApp:
    memorizations: list[Memorization | Memorization2]

    def save(self) -> None:
        """
        Write it all out to somewhere persistent.
        """
        with open(timecache, "w") as f:
            f.write(
                str(
                    min([each.nextPromptTime() for each in self.memorizations])
                    if self.memorizations
                    else 0
                )
            )
        set_password(
            "pinpal", "storage", dumps([each.tojson() for each in self.memorizations])
        )

    @classmethod
    def load(cls) -> PinPalApp | None:
        """
        Load it from somewhere persistent.
        """
        stored = get_password("pinpal", "storage")
        if stored is None:
            return None
        self = PinPalApp([load(each) for each in loads(stored)])
        return self


def load(x: dict[str, object]) -> Memorization | Memorization2:
    if "targetTokenCount" in x:
        return Memorization2.fromjson(x)
    else:
        return Memorization.fromjson(x)





"""
This is a setup.py script generated by py2applet

Usage:
    python py2app_setup.py py2app
"""

import py2app
import os

from setuptools import setup

def check_mode() -> str:
    match os.environ:
        case {"CI_MODE": b}:
            if b:
                return "Ci"
        case {"TEST_MODE": b}:
            if b:
                return "Test"
    return ""

MODE = check_mode()

APP = [f"PINPalMac/{MODE}PINPal.py"]
DATA_FILES = [
    "PINPalMac/MainMenu.xib",
    "PINPalMac/PINList.xib",
]
OPTIONS = {
    "plist": {
        "LSUIElement": True,
        "NSRequiresAquaSystemAppearance": False,
        "CFBundleIdentifier": f"im.glyph.and.this.is.{MODE}pinpal",
        "CFBundleName": f"{MODE}PINPal",
    },
    "iconfile": f"{MODE}icon.icns",
    "app": APP,
}

setup(
    data_files=DATA_FILES,
    options={"py2app": OPTIONS},
)
